/*****************************************************************************

        main.cpp
        Copyright (c) 2005 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (4 : 4786) // "identifier was truncated to '255' characters in the debug information"
	#pragma warning (4 : 4800) // "forcing value to bool 'true' or 'false' (performance warning)"
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"ClockCycleCounter.h"

#if defined (_MSC_VER)
#include	<crtdbg.h>
#include	<new.h>
#endif	// _MSC_VER

#include	<new>

#include	<cassert>
#include	<cmath>
#include	<cstdio>



static void	display_results (const stopwatch::ClockCycleCounter &ccc, long nbr_it);
static void	test_empty_loop (stopwatch::ClockCycleCounter &ccc);
static void	test_sine_sum (stopwatch::ClockCycleCounter &ccc);



/*\\\ FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



#if defined (_MSC_VER)
static int __cdecl	STPW_new_handler_cb (size_t dummy)
{
	throw std::bad_alloc ();
	return (0);
}
#endif	// _MSC_VER



#if defined (_MSC_VER) && ! defined (NDEBUG)
static int	__cdecl	STPW_debug_alloc_hook_cb (int alloc_type, void *user_data_ptr, size_t size, int block_type, long request_nbr, const unsigned char *filename_0, int line_nbr)
{
	if (block_type != _CRT_BLOCK)	// Ignore CRT blocks to prevent infinite recursion
	{
		switch (alloc_type)
		{
		case	_HOOK_ALLOC:
		case	_HOOK_REALLOC:
		case	_HOOK_FREE:

			// Put some debug code here

			break;

		default:
			assert (false);	// Undefined allocation type
			break;
		}
	}

	return (1);
}
#endif



#if defined (_MSC_VER) && ! defined (NDEBUG)
static int	__cdecl	STPW_debug_report_hook_cb (int report_type, char *user_msg_0, int *ret_val_ptr)
{
	*ret_val_ptr = 0;	// 1 to override the CRT default reporting mode

	switch (report_type)
	{
	case	_CRT_WARN:
	case	_CRT_ERROR:
	case	_CRT_ASSERT:

		// Put some debug code here

		break;
	}

	return (*ret_val_ptr);
}
#endif



static void	STPW_main_prog_init ()
{
#if defined (_MSC_VER)
	::_set_new_handler (::STPW_new_handler_cb);
#endif	// _MSC_VER

#if defined (_MSC_VER) && ! defined (NDEBUG)
	{
		const int	mode =   (1 * _CRTDBG_MODE_DEBUG)
						       | (1 * _CRTDBG_MODE_WNDW);
		::_CrtSetReportMode (_CRT_WARN, mode);
		::_CrtSetReportMode (_CRT_ERROR, mode);
		::_CrtSetReportMode (_CRT_ASSERT, mode);

		const int	old_flags = ::_CrtSetDbgFlag (_CRTDBG_REPORT_FLAG);
		::_CrtSetDbgFlag (  old_flags
		                  | (1 * _CRTDBG_LEAK_CHECK_DF)
		                  | (1 * _CRTDBG_CHECK_ALWAYS_DF));
		::_CrtSetBreakAlloc (-1);	// Specify here a memory bloc number
		::_CrtSetAllocHook (STPW_debug_alloc_hook_cb);
		::_CrtSetReportHook (STPW_debug_report_hook_cb);

		// Speed up I/O but breaks C stdio compatibility
//		std::cout.sync_with_stdio (false);
//		std::cin.sync_with_stdio (false);
//		std::cerr.sync_with_stdio (false);
//		std::clog.sync_with_stdio (false);
	}
#endif	// _MSC_VER, NDEBUG
}



static void	STPW_main_prog_end ()
{
#if defined (_MSC_VER) && ! defined (NDEBUG)
	{
		const int	mode =   (1 * _CRTDBG_MODE_DEBUG)
						       | (0 * _CRTDBG_MODE_WNDW);
		::_CrtSetReportMode (_CRT_WARN, mode);
		::_CrtSetReportMode (_CRT_ERROR, mode);
		::_CrtSetReportMode (_CRT_ASSERT, mode);

		::_CrtMemState	mem_state;
		::_CrtMemCheckpoint (&mem_state);
		::_CrtMemDumpStatistics (&mem_state);
	}
#endif	// _MSC_VER, NDEBUG
}



int main (int argc, char *argv [])
{
	using namespace std;

	STPW_main_prog_init ();

	try
	{
		stopwatch::ClockCycleCounter	ccc;

		// Measures an empty loop
		test_empty_loop (ccc);

		// Measures a sum of sines
		test_sine_sum (ccc);
	}

	catch (std::exception &e)
	{
		printf ("*** main(): Exception (std::exception): %s\n", e.what ());
		throw;
	}

	catch (...)
	{
		printf ("*** main(): Undefined exception.\n");
		throw;
	}

	STPW_main_prog_end ();

	return (0);
}



void	display_results (const stopwatch::ClockCycleCounter &ccc, long nbr_it, long nbr_it_per_lap)
{
	assert (&ccc != 0);
	assert (nbr_it > 0);

	const double	total = static_cast <double> (ccc.get_time_total ());

	using namespace std;

	printf ("Total (%ld iterations): %.1f clock cycles.\n", nbr_it, total);
	printf ("Iterations duration (raw): %.1f clock cycles.\n", total / nbr_it);
	if (nbr_it_per_lap == 1)
	{
		printf (
			"Iterations duration (net): %.1f clock cycles.\n",
			static_cast <double> (ccc.get_time_best_lap ()) / nbr_it_per_lap
		);
	}
}



void	test_empty_loop (stopwatch::ClockCycleCounter &ccc)
{
	assert (&ccc != 0);

	using namespace std;

	printf ("Measuring empty loop duration...\n");
	const long		nbr_loops = 10000000;
	ccc.start ();
	for (long cnt = 0; cnt < nbr_loops; ++cnt)
	{
		ccc.stop_lap ();
	}

	display_results (ccc, nbr_loops, 1);
	printf ("\n");
}



void	test_sine_sum (stopwatch::ClockCycleCounter &ccc)
{
	assert (&ccc != 0);

	using namespace std;

	const long		nbr_loops = 10000;
	long				cnt;
	double			sum;
	double			phase ;

	// With lap
	printf ("Measuring a sum of sines...\n");
	sum = 0;
	phase = 0;
	ccc.start ();
	for (cnt = 0; cnt < nbr_loops; ++cnt)
	{
		for (int k = 0; k < 100; ++k)
		{
			sum += sin (phase);
			sum += sin (phase + 3.1415927*0.5);
			phase += 1e-6;
		}
		ccc.stop_lap ();
	}
	display_results (ccc, nbr_loops, 1);
	printf ("Sum: %f\n", sum);

	// Without lap
	printf ("And withoug lap measure...\n");
	sum = 0;
	phase = 0;
	ccc.start ();
	for (cnt = 0; cnt < nbr_loops; ++cnt)
	{
		for (int k = 0; k < 100; ++k)
		{
			sum += sin (phase);
			sum += sin (phase + 3.1415927*0.5);
			phase += 1e-6;
		}
	}
	ccc.stop_lap ();
	display_results (ccc, nbr_loops, nbr_loops);
	printf ("Sum: %f\n\n", sum);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
